#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/i2c.h>
#include <linux/types.h>
#include <linux/videodev.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/slab.h>

#include "id.h"
#include "audiochip.h"

/* Addresses to scan */
static unsigned short normal_i2c[] = {I2C_CLIENT_END};
static unsigned short normal_i2c_range[] = {0x86>>1,0x86>>1,I2C_CLIENT_END};
static unsigned short probe[2]        = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short probe_range[2]  = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short ignore[2]       = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short ignore_range[2] = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short force[2]        = { I2C_CLIENT_END, I2C_CLIENT_END };
static struct i2c_client_address_data addr_data = {
	normal_i2c, normal_i2c_range, 
	probe, probe_range, 
	ignore, ignore_range, 
	force
};

/* insmod options */
static int debug =  0;
static char *pal =  "b";
MODULE_PARM(debug,"i");
MODULE_PARM(pal,"s");
MODULE_LICENSE("GPL");

struct tda9887 {
	struct i2c_client client;
	int radio,tvnorm;
};

static struct i2c_driver driver;
static struct i2c_client client_template;

/* ---------------------------------------------------------------------- */

static unsigned char buf_pal_bg[]    = { 0x00, 0x16, 0x70, 0x49 };
static unsigned char buf_pal_i[]     = { 0x00, 0x16, 0x70, 0x4a };
static unsigned char buf_pal_dk[]    = { 0x00, 0x16, 0x70, 0x4b };
static unsigned char buf_pal_l[]     = { 0x00, 0x06, 0x50, 0x4b };
static unsigned char buf_fm_stereo[] = { 0x00, 0x0e, 0x0d, 0x77 };

static int tda9887_configure(struct tda9887 *t)
{
	unsigned char *buf = NULL;
	int rc;

	if (t->radio)
		buf = buf_fm_stereo;
	else if (t->tvnorm == VIDEO_MODE_PAL) {
		switch (pal[0]) {
		case 'b':
		case 'g':
			buf = buf_pal_bg;
			break;
		case 'i':
			buf = buf_pal_i;
			break;
		case 'd':
		case 'k':
			buf = buf_pal_dk;
			break;
		case 'l':
			buf = buf_pal_l;
			break;
		}
	}
	if (NULL == buf)
		return 0;

	if (debug)
		printk("tda9887: 0x%02x 0x%02x 0x%02x 0x%02x\n",
		       buf[0],buf[1],buf[2],buf[3]);
        if (4 != (rc = i2c_master_send(&t->client,buf,4)))
                printk("tda9887: i2c i/o error: rc == %d (should be 4)\n",rc);
	return 0;
}

/* ---------------------------------------------------------------------- */

static int tda9887_attach(struct i2c_adapter *adap, int addr,
			  unsigned short flags, int kind)
{
	struct tda9887 *t;

        client_template.adapter = adap;
        client_template.addr    = addr;

        printk("tda9887: chip found @ 0x%x\n", addr<<1);

        if (NULL == (t = kmalloc(sizeof(*t), GFP_KERNEL)))
                return -ENOMEM;
	memset(t,0,sizeof(*t));
	t->client = client_template;
        t->client.data = t;
        i2c_attach_client(&t->client);
        
	MOD_INC_USE_COUNT;
	return 0;
}

static int tda9887_probe(struct i2c_adapter *adap)
{
	int rc;

	switch (adap->id) {
	case I2C_ALGO_BIT | I2C_HW_B_BT848:
	case I2C_ALGO_BIT | I2C_HW_B_RIVA:
	case I2C_ALGO_SAA7134:
		printk("tda9887: probing %s i2c adapter [id=0x%x]\n",
		       adap->name,adap->id);
		rc = i2c_probe(adap, &addr_data, tda9887_attach);
		break;
	default:
		printk("tda9887: ignoring %s i2c adapter [id=0x%x]\n",
		       adap->name,adap->id);
		rc = 0;
		/* nothing */
	}
	return rc;
}

static int tda9887_detach(struct i2c_client *client)
{
	struct tda9887 *t = (struct tda9887*)client->data;

	i2c_detach_client(client);
	kfree(t);
	MOD_DEC_USE_COUNT;
	return 0;
}

static int
tda9887_command(struct i2c_client *client, unsigned int cmd, void *arg)
{
	struct tda9887 *t = (struct tda9887*)client->data;

        switch (cmd) {

	/* --- configuration --- */
	case AUDC_SET_RADIO:
		t->radio = 1;
		tda9887_configure(t);
		break;
		
	/* --- v4l ioctls --- */
	/* take care: bttv does userspace copying, we'll get a
	   kernel pointer here... */
	case VIDIOCSCHAN:
	{
		struct video_channel *vc = arg;

		t->radio  = 0;
		t->tvnorm = vc->norm;
		tda9887_configure(t);
		break;
	}
	default:
		/* nothing */
		break;
	}
	return 0;
}

/* ----------------------------------------------------------------------- */

static struct i2c_driver driver = {
        name:           "i2c tda9887 driver",
        id:             -1, /* FIXME */
        flags:          I2C_DF_NOTIFY,
        attach_adapter: tda9887_probe,
        detach_client:  tda9887_detach,
        command:        tda9887_command,
};
static struct i2c_client client_template =
{
        name:   "tda9887",
	flags:  I2C_CLIENT_ALLOW_USE,
        driver: &driver,
};

static int tda9887_init_module(void)
{
	i2c_add_driver(&driver);
	return 0;
}

static void tda9887_cleanup_module(void)
{
	i2c_del_driver(&driver);
}

module_init(tda9887_init_module);
module_exit(tda9887_cleanup_module);

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
